/***************************************************************/
/*     Performance Analysis in CUDA                            */
/*                                                             */
/*                                                             */
/*     Author: E. Bartocci (eziobart@ams.sunysb.edu)           */
/*                                                             */
/*     Description:                                            */
/*                                                             */
/*        Just a simple application in order to measure        */
/*        performances                                         */
/*                                                             */ 
/*     Date:  11/29/10                                         */
/*                                                             */
/*   Free distribution with authors permission                 */
/*                                                             */
/* ==   SUNY Stony Brook, Stony Brook, NY                      */
/* ==                                                          */ 
/* ============================================================*/   

/*Exercises: */

/* 1) Changing the BLOCK_SIZE and observe what will happen
 * 2) Changing the MATRIX_SIZE and observe what will happen
 * 3) What is the maximum theretical speedup using Amdahl's Law ?
 * 4) Measuring the cost of communication between the CPU and GPU 
 */


#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <cutil_inline.h>

#define MATRIX_SIZE matrix_size[mi]
#define BLOCK_SIZE block_size[bi]

__global__ void add_matrix (float* a, float *b, float *c, int N){
           int i = blockIdx.x * blockDim.x + threadIdx.x;
           int j = blockIdx.y * blockDim.y + threadIdx.y;
           int index = i + j*N;
           if (i < N && j <  N)
               c[index] = a[index] + b[index];
}

int main(){
	int i, j;
	float *A, *B;

	float *Ad, *Bd;
	clock_t c1, c2;
	double serial_io = 0, serial_comp = 0;
	int matrix_size [] = {512, 1024, 2048, 4096, 8192, 16384};
	int block_size [] = {2, 4, 8, 16, 32};
	int mi, bi;
printf("Matrix Size,Block Size, Device MP, Device SP, Problem size, Processes, Serial I/O time, Serial Calc time, Serial Total, Parallel I/O time, Parallel Calc, Parallel Total, SpeedupCalc, SpeedUpTotal, EfficiencyCalc, EfficiencyTotal\n");
	for (mi=0; mi < 6; mi++){
		for (bi=0; bi < 5; bi++){
        // get number of SMs on this GPU
	int devID;
    cudaDeviceProp props;
	
    cutilSafeCall(cudaGetDevice(&devID));
	cutilSafeCall(cudaGetDeviceProperties(&props, devID));

	/*Allocation of space for the Matrix*/

	c1 = clock();
	A = (float*) malloc(MATRIX_SIZE * MATRIX_SIZE * sizeof(float));
	B = (float*) malloc(MATRIX_SIZE * MATRIX_SIZE * sizeof(float));
	
	for(j = 0; j < MATRIX_SIZE; j++){
		for(i = 0; i < MATRIX_SIZE; i++){
			A[i * MATRIX_SIZE + j]= (i + 1) + (j + 1) * 1024;
		}
	}
	c2 = clock();
	
	serial_io = c2 - c1;
	
	c1 = clock();
	for(j = 0; j < MATRIX_SIZE; j++){
			for(i = 0; i < MATRIX_SIZE; i++){
				B[i * MATRIX_SIZE + j]= A[j * MATRIX_SIZE + i] + A[j * MATRIX_SIZE + i];
			}
        }
	c2 = clock();
	serial_comp = c2 - c1;
	free(A);
	free(B);
	double parallel_io = 0, parallel_comp = 0;
	
	c1 = clock();    
	A = (float*) malloc(MATRIX_SIZE * MATRIX_SIZE * sizeof(float));
	B = (float*) malloc(MATRIX_SIZE * MATRIX_SIZE * sizeof(float));
	
	for(j = 0; j < MATRIX_SIZE; j++){
		for(i = 0; i < MATRIX_SIZE; i++){
			A[i * MATRIX_SIZE + j]= (i + 1) + (j + 1) * 1024;
		}
	}
	cudaMalloc( (void**)&Ad, MATRIX_SIZE*MATRIX_SIZE*sizeof(float));
	cudaMalloc( (void**)&Bd, MATRIX_SIZE*MATRIX_SIZE*sizeof(float));
    cudaMemcpy( Ad, A, MATRIX_SIZE*MATRIX_SIZE*sizeof(float), cudaMemcpyHostToDevice );
    cudaMemcpy( Bd, B, MATRIX_SIZE*MATRIX_SIZE*sizeof(float), cudaMemcpyHostToDevice );
    
    c2 = clock();
    parallel_io = c2 - c1;
    
	c1 = clock();
	dim3 dimBlock( BLOCK_SIZE, BLOCK_SIZE );
    dim3 dimGrid( MATRIX_SIZE/dimBlock.x, MATRIX_SIZE/dimBlock.y );
	
    add_matrix<<<dimGrid, dimBlock>>>( Ad, Ad, Bd, MATRIX_SIZE );
	cudaMemcpy( B, Bd, MATRIX_SIZE, cudaMemcpyDeviceToHost );
	cutilSafeCall( cudaThreadSynchronize() );
	c2 = clock();
	parallel_comp = c2 - c1;
	
	        printf("%d,", MATRIX_SIZE );
	        printf("%d,", BLOCK_SIZE );
    printf("%d, %d, ", props.multiProcessorCount, props.multiProcessorCount*8 );
    printf(" %d,", MATRIX_SIZE*MATRIX_SIZE);
    printf("%d,", props.multiProcessorCount*8);
	printf(" %.0f,", serial_io / (CLOCKS_PER_SEC / 1000));
	printf("%.0f,", serial_comp / (CLOCKS_PER_SEC / 1000));
	
	printf("%.0f,", (serial_io + serial_comp)  / (CLOCKS_PER_SEC / 1000));
	printf("%.0f,", parallel_comp / (CLOCKS_PER_SEC / 1000));
	printf("%.0f,", parallel_io / (CLOCKS_PER_SEC / 1000));
	
	printf("%.0f,", (parallel_io + parallel_comp)  / (CLOCKS_PER_SEC / 1000));
	printf(" %.4f,",  serial_comp/parallel_comp);
	printf("%.4f,",   (serial_io+serial_comp)/(parallel_io+parallel_comp));
	printf("%.4f,",  (serial_comp)/(props.multiProcessorCount*8 * (parallel_comp)));
	printf("%.4f\n",  (serial_io + serial_comp)/(props.multiProcessorCount*8 * (parallel_io + parallel_comp)));

	//printf("First line of matrix:");
	//for ( int i = 0; i < 5; ++i ) {printf("%f %f", A[i], B[i]); }
	//printf("\n");
	cudaFree( Ad );                                                              
	cudaFree( Bd );
	free(A);
	free(B);
		}
	}
}


